.flatten-tab-title {
    font-weight: bold;
    color: blue;
}

.flatten-sphinx-tabs-tab {
  position: relative;
  font-family: Lato,'Helvetica Neue',Arial,Helvetica,sans-serif;
  color: #1D5C87;
  line-height: 24px;
  margin: 0;
  font-size: 16px;
  font-weight: 400;
  background-color: rgba(255, 255, 255, 0);
  border-radius: 5px 5px 5px 5px;
  border: 1px solid ;
  padding: 1rem 1.5rem;
  margin-bottom: 0.5rem;
}